import tomllib
import warnings
from pathlib import Path

import geopandas as gpd
import geost
import rioxarray as rio
import typer
import xarray as xr
from geost.validate.validate import ValidationWarning
from shapely import geometry as gmt

from geosections import base, utils

warnings.filterwarnings("ignore", category=ValidationWarning)


def _geopandas_read(file: str | Path, **kwargs) -> gpd.GeoDataFrame:
    file = Path(file)
    if file.suffix in {".shp", ".gpkg"}:
        return gpd.read_file(file, **kwargs)
    elif file.suffix in {".parquet", ".geoparquet"}:
        return gpd.read_parquet(file, **kwargs)
    else:
        raise ValueError(f"File type {file.suffix} is not supported by geopandas.")


def read_config(file: str | Path) -> base.Config:
    """
    Read a TOML configuration file and return a Config object for `geosections` tools.

    Parameters
    ----------
    file : str | Path
        Pathlike object to the TOML configuration file.

    Returns
    -------
    :class:`~geosections.Config`
        Configuration object for `geosections` tools.

    """
    with open(file, "rb") as f:
        config = tomllib.load(f)

    try:
        config = base.Config(**config)
    except Exception as e:
        typer.secho(f"Invalid configuration:\n{e}", fg=typer.colors.RED)
        raise typer.Exit(code=1)

    return config


def read_line(data: base.Line) -> gmt.LineString:
    line = _geopandas_read(data.file)

    if line.crs is None or line.crs != 28992:
        line.set_crs(28992, allow_override=True, inplace=True)

    if data.name is not None:
        try:
            line = line[line[data.name_column] == data.name]["geometry"].iloc[0]
        except KeyError as e:
            typer.secho(
                f"'name_column' not found in input cross-section lines:\n{e}",
                fg=typer.colors.RED,
            )
            raise typer.Exit(code=1)
    else:
        line = line["geometry"].iloc[0]

    return line


def read_boreholes(
    data: base.Data, line: gmt.LineString
) -> geost.base.BoreholeCollection:
    boreholes = geost.read_borehole_table(data.file, horizontal_reference=data.crs)

    if boreholes.horizontal_reference != 28992:
        boreholes.change_horizontal_reference(28992)

    boreholes = boreholes.select_with_lines(line, buffer=data.max_distance_to_line)
    boreholes.header["dist"] = utils.distance_on_line(boreholes, line)
    return boreholes


def read_cpts(data: base.Data, line: gmt.LineString) -> geost.base.BoreholeCollection:
    cpts = geost.read_cpt_table(data.file, horizontal_reference=data.crs)

    if cpts.horizontal_reference != 28992:
        cpts.change_horizontal_reference(28992)

    cpts = utils.cpts_to_borehole_collection(
        cpts.select_with_lines(line, buffer=30),
        {
            "depth": ["min", "max"],
            "lith": "first",
        },
    )
    cpts.header["dist"] = utils.distance_on_line(cpts, line)
    cpts.add_header_column_to_data("surface")
    cpts.add_header_column_to_data("end")
    return cpts


def read_surface(data: base.Surface, line: gmt.LineString) -> xr.DataArray:
    surface = rio.open_rasterio(data.file, masked=True).squeeze(drop=True)

    if surface.rio.crs is None:
        warning = (
            f"Surface {Path(data.file).stem} has no CRS, surface may not be shown correctly "
            "along the cross-section line."
        )
        typer.secho(warning, fg=typer.colors.YELLOW)
    elif surface.rio.crs != 28992:
        surface = surface.rio.reproject(28992)

    surface = geost.models.model_utils.sample_along_line(surface, line, dist=2.5)
    return surface


def read_curves(config: base.Config, line: gmt.LineString) -> geost.base.CptCollection:
    curves = geost.read_cpt_table(
        config.data.cpts.file, horizontal_reference=config.data.cpts.crs
    )

    if curves.horizontal_reference != 28992:
        curves.change_horizontal_reference(28992)

    curves = utils.get_cpt_curves_for_section(
        curves,
        config.data.curves.nrs,
        line,
        dist_scale_factor=config.data.curves.dist_scale_factor,
    )
    return curves



def plot_borehole_data(ax, data, colors, width=20):
    for nr, dist in zip(data.header["nr"], data.header["dist"]):
        c = data.data[data.data["nr"] == nr]
        plot_borehole(ax, c, dist, width, colors)
    return


def plot_borehole(ax, df, dist, width, colors):
    df["top"] = df["surface"] - df["top"]
    df["bottom"] = df["surface"] - df["bottom"]

    thickness = df["top"] - df["bottom"]
    for lith, t, bot in zip(df["lith"], thickness, df["bottom"]):
        ax.bar(dist, t, bottom=bot, color=colors.get(lith, "grey"), width=width)
    return


def plot_curves(ax, curves, ymax):
    for nr in curves.header["nr"]:
        c = curves.get(nr)
        ax.plot(c.data["qc"], c.data["depth"], color="r", linewidth=0.5)
        ax.plot(c.data["fs"], c.data["depth"], color="b", linewidth=0.5)
        ax.text(
            c.header["dist"],
            ymax + 2,
            nr.replace("000000", ""),
            rotation=90,
            ha="center",
            va="bottom",
            fontsize=8,
        )
    return



import geost
import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

type BoreholeCollection = geost.base.BoreholeCollection
type CptCollection = geost.base.CptCollection
type Collection = geost.base.Collection


def cpts_to_borehole_collection(
    cpts: CptCollection, aggfuncs: dict
) -> BoreholeCollection:
    _, lith_nrs = np.unique(cpts.data["lith"], return_inverse=True)
    cpts.data["lith_nr"] = lith_nrs

    cpts_top_bot = cpts_as_top_bottom(cpts, aggfuncs, "lith_nr")
    cpts_top_bot.rename(
        columns={"depth_min": "top", "depth_max": "bottom"}, inplace=True
    )
    layered = geost.base.LayeredData(cpts_top_bot, has_inclined=False)
    result = geost.base.BoreholeCollection(cpts.header, layered)
    return result


def cpts_as_top_bottom(
    cpts: CptCollection, aggfuncs: dict, layer_col: str
) -> pd.DataFrame:
    data = cpts.data.df.copy()
    data["layer"] = create_layer_numbers(cpts, layer_col)

    cpts_as_top_bot = pd.pivot_table(
        data,
        index=["nr", "layer"],
        aggfunc=aggfuncs,
        sort=False,
    )

    cpts_as_top_bot.columns = _get_columns(aggfuncs)
    cpts_as_top_bot.reset_index(inplace=True)
    return cpts_as_top_bot


def _get_columns(aggfuncs: dict):
    for key, value in aggfuncs.items():
        if isinstance(value, str):
            yield key
        else:
            for v in value:
                yield f"{key}_{v}"


def create_layer_numbers(collection: Collection, layer_col: str) -> pd.Series:
    return (
        collection.data.df.groupby("nr")
        .apply(
            lambda x: label_consecutive_elements(x[layer_col].values),
            include_groups=False,
        )
        .explode()
        .values
    )


def label_consecutive_elements(array: np.ndarray) -> np.ndarray:
    """
    Label consecutive elements in an array.

    Parameters:
    -----------
    array : np.ndarray
        The array to label.

    Returns:
    --------
    np.ndarray
        The labeled array.

    """
    diff = np.diff(array, prepend=array[0])
    return np.cumsum(diff != 0)


def distance_on_line(collection, line):
    return line.project(collection.header["geometry"])


def get_cpt_curves_for_section(cpt_data, nrs, line, dist_scale_factor=80):
    cpt_curves = cpt_data.get(nrs)
    cpt_curves.header["dist"] = line.project(cpt_curves.header.gdf["geometry"])

    scaler = MinMaxScaler()
    cpt_curves.data["qc"] = (
        scaler.fit_transform(cpt_curves.data["cone_resistance"].values.reshape(-1, 1))
        * dist_scale_factor
    )
    cpt_curves.data["fs"] = (
        scaler.fit_transform(cpt_curves.data["friction_ratio"].values.reshape(-1, 1))
        * dist_scale_factor
    )
    cpt_curves.data["depth"] = cpt_curves.data["surface"] - cpt_curves.data["depth"]
    cpt_curves.add_header_column_to_data("dist")
    cpt_curves.data["fs"] *= -1
    cpt_curves.data["qc"] += cpt_curves.data["dist"]
    cpt_curves.data["fs"] += cpt_curves.data["dist"]
    return cpt_curves



from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field


class PlotLabels(BaseModel):
    xlabel: str = Field("")
    ylabel: str = Field("")
    title: str = Field("")


class PlotSettings(BaseModel):
    column_with: int | float = Field(default=20)
    fig_width: int = Field(default=11)
    fig_height: int = Field(default=7)
    inches: bool = Field(default=True)
    grid: bool = Field(default=True)
    dpi: int = Field(default=300)
    tight_layout: bool = Field(default=True)
    ymin: int | float = Field(default=None)
    ymax: int | float = Field(default=None)


class Surface(BaseModel):
    file: str
    style_kwds: dict[str, Any] = Field(default={})


class Data(BaseModel):
    file: str
    max_distance_to_line: int | float = Field(default=50)
    crs: int = Field(default=28992)


class Curves(BaseModel):
    nrs: list[str]
    dist_scale_factor: int | float = Field(default=80)


class PlotData(BaseModel):
    boreholes: Data = Field(default=None)
    cpts: Data = Field(default=None)
    curves: Curves = Field(default=None)


class Line(BaseModel):
    file: str | Path
    crs: int = Field(default=28992)
    name: Any = Field(default=None)
    name_column: str = Field(default="name")


class Config(BaseModel):
    line: Line
    data: PlotData = Field(default=PlotData())
    surface: list[Surface] = Field(default=[])
    labels: PlotLabels = Field(default=PlotLabels())
    settings: PlotSettings = Field(default=PlotSettings())
    colors: dict[str, str] = Field({"default": "#000000"})



import typer
from matplotlib import pyplot as plt

from geosections import plotting, read

app = typer.Typer()


@app.command()
def plot(
    config: str = typer.Argument(..., help="Path to TOML-configuration file"),
    output_file: str = typer.Option(None, "--save", help="Path to output file"),
    close: bool = typer.Option(False, "--close", help="Close plot"),
):
    config = read.read_config(config)
    line = read.read_line(config.line)

    fig, ax = plt.subplots(
        figsize=(config.settings.fig_width, config.settings.fig_height),
        tight_layout=config.settings.tight_layout,
    )

    if config.data.boreholes is not None:
        typer.secho(
            f"Plotting boreholes from {config.data.boreholes.file}",
            fg=typer.colors.BLUE,
        )
        boreholes = read.read_boreholes(config.data.boreholes, line)
        plotting.plot_borehole_data(
            ax, boreholes, config.colors, config.settings.column_with
        )

    if config.data.cpts is not None:
        typer.secho(
            f"Plotting CPTs from {config.data.cpts.file}",
            fg=typer.colors.BLUE,
        )
        cpts = read.read_cpts(config.data.cpts, line)
        plotting.plot_borehole_data(
            ax, cpts, config.colors, config.settings.column_with
        )

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin - 2, ymax)
    ax.set_xlim(0, line.length)
    ax.set_xlabel(config.labels.xlabel)
    ax.set_ylabel(config.labels.ylabel)
    ax.set_title(config.labels.title)
    ax.grid(config.settings.grid, linestyle="--", alpha=0.5)

    if config.data.curves is not None:
        typer.secho(
            f"Plotting curves from {config.data.curves.nrs}",
            fg=typer.colors.BLUE,
        )
        curves = read.read_curves(config, line)
        plotting.plot_curves(ax, curves, ymax)

    if config.surface:
        for surface in config.surface:
            typer.secho(
                f"Plotting surface from {surface.file}",
                fg=typer.colors.BLUE,
            )
            surface_line = read.read_surface(surface, line)
            ax.plot(
                surface_line["dist"].values, surface_line.values, **surface.style_kwds
            )

    if output_file:
        fig.savefig(output_file)

    if close:
        plt.close()
    else:
        plt.show()


@app.command()
def check_unique_lithologies(
    config: str = typer.Argument(..., help="Pad naar TOML-configuratiebestand")
):
    config = read.read_config(config)
    line = read.read_line(config.line)
    boreholes = read.read_boreholes(config.data.boreholes, line)
    cpts = read.read_cpts(config.data.cpts, line)

    uniques = set(boreholes.data["lith"]) | set(cpts.data["lith"])
    typer.secho(
        f"Unique lithologies in boreholes: {uniques}",
        fg=typer.colors.BLUE,
    )
